from typing import List, Optional

from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import Reference, SkosCode, SkosConcept
from fdk_rdf_parser.rdf_utils import value_list


def extract_references(graph: Graph, subject: URIRef) -> Optional[List[Reference]]:
    reference_properties = [
        DCTERMS.hasVersion,
        DCTERMS.isVersionOf,
        DCTERMS.isPartOf,
        DCTERMS.hasPart,
        DCTERMS.references,
        DCTERMS.isReferencedBy,
        DCTERMS.replaces,
        DCTERMS.isReplacedBy,
        DCTERMS.requires,
        DCTERMS.isRequiredBy,
        DCTERMS.relation,
        DCTERMS.source,
    ]

    values = []

    for predicate in reference_properties:
        references = value_list(graph, subject, predicate)
        if references:
            for reference in references:
                values.append(
                    Reference(
                        referenceType=SkosCode(uri=predicate.toPython()),
                        source=SkosConcept(uri=reference),
                    )
                )

    return values if len(values) > 0 else None



from dataclasses import dataclass
from typing import Dict, List, Optional

from .skos_concept import SkosConcept


@dataclass
class CriterionRequirement:
    uri: Optional[str] = None
    identifier: Optional[str] = None
    name: Optional[Dict[str, str]] = None
    type: Optional[List[SkosConcept]] = None



from typing import List

from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS, FOAF, OWL

from fdk_rdf_parser.classes import InformationModel
from fdk_rdf_parser.rdf_utils import (
    adms_uri,
    catalog_ref,
    model_dcat_ap_no_uri,
    object_value,
    resource_list,
    uri_or_identifier,
    uri_or_identifier_list,
    value_set,
)
from .catalog import parse_catalog
from .dcat_resource import parse_dcat_resource
from .harvest_meta_data import extract_meta_data
from .model_element import parse_model_element
from .model_property import parse_model_property
from .skos_code import extract_skos_code_list
from .temporal import extract_temporal


def parse_information_model(
    graph: Graph, fdk_record_uri: URIRef, info_model_uri: URIRef
) -> InformationModel:

    subjects = value_set(graph, info_model_uri, DCTERMS.subject)
    model_element_refs = resource_list(
        graph, info_model_uri, model_dcat_ap_no_uri("containsModelElement")
    )

    info_model = InformationModel(
        id=object_value(graph, fdk_record_uri, DCTERMS.identifier),
        harvest=extract_meta_data(graph, fdk_record_uri),
        catalog=parse_catalog(graph, fdk_record_uri),
        license=extract_skos_code_list(graph, info_model_uri, DCTERMS.license),
        informationModelIdentifier=object_value(
            graph, info_model_uri, model_dcat_ap_no_uri("informationModelIdentifier")
        ),
        spatial=extract_skos_code_list(graph, info_model_uri, DCTERMS.spatial),
        temporal=extract_temporal(graph, info_model_uri),
        isPartOf=object_value(graph, info_model_uri, DCTERMS.isPartOf),
        hasPart=object_value(graph, info_model_uri, DCTERMS.hasPart),
        isReplacedBy=object_value(graph, info_model_uri, DCTERMS.isReplacedBy),
        replaces=object_value(graph, info_model_uri, DCTERMS.replaces),
        hasFormat=value_set(graph, info_model_uri, DCTERMS.hasFormat),
        homepage=object_value(graph, info_model_uri, FOAF.homepage),
        status=object_value(graph, info_model_uri, adms_uri("status")),
        versionInfo=object_value(graph, info_model_uri, OWL.versionInfo),
        versionNotes=object_value(graph, info_model_uri, OWL.versionNotes),
        subjects=subjects,
        containsSubjects=subjects.copy() if subjects else None,
        containsModelElements=uri_or_identifier_list(graph, model_element_refs),
    )

    info_model.add_values_from_dcat_resource(
        parse_dcat_resource(
            graph,
            info_model_uri,
            catalog_subject=catalog_ref(graph, fdk_record_uri),
        )
    )

    return add_elements_to_model(info_model, graph, model_element_refs)


def add_elements_to_model(
    info_model: InformationModel, graph: Graph, element_refs: List[URIRef]
) -> InformationModel:
    for element_ref in element_refs:

        element_id_optional = uri_or_identifier(graph, element_ref)
        if element_id_optional:
            element_id = str(element_id_optional)
            if (
                info_model.modelElements is None
                or info_model.modelElements.get(element_id) is None
            ):
                info_model.add_model_element(parse_model_element(graph, element_ref))

                has_property_refs = resource_list(
                    graph, element_ref, model_dcat_ap_no_uri("hasProperty")
                )
                info_model = add_properties_to_model(
                    info_model, graph, has_property_refs
                )

    return info_model


def add_properties_to_model(
    info_model: InformationModel, graph: Graph, prop_refs: List[URIRef]
) -> InformationModel:
    for prop_ref in prop_refs:

        prop_id_optional = uri_or_identifier(graph, prop_ref)
        if prop_id_optional:
            prop_id = str(prop_id_optional)
            if (
                info_model.modelProperties is None
                or info_model.modelProperties.get(prop_id) is None
            ):
                info_model.add_model_property(parse_model_property(graph, prop_ref))

                property_ref_predicates = [model_dcat_ap_no_uri("formsSymmetryWith")]

                for predicate in property_ref_predicates:
                    property_refs = resource_list(graph, prop_ref, predicate)
                    info_model = add_properties_to_model(
                        info_model, graph, property_refs
                    )

                element_ref_predicates = [
                    model_dcat_ap_no_uri("hasType"),
                    model_dcat_ap_no_uri("isAbstractionOf"),
                    model_dcat_ap_no_uri("refersTo"),
                    model_dcat_ap_no_uri("hasDataType"),
                    model_dcat_ap_no_uri("hasSimpleType"),
                    model_dcat_ap_no_uri("hasObjectType"),
                    model_dcat_ap_no_uri("hasValueFrom"),
                    model_dcat_ap_no_uri("hasSome"),
                    model_dcat_ap_no_uri("hasMember"),
                    model_dcat_ap_no_uri("contains"),
                    model_dcat_ap_no_uri("hasSupplier"),
                    model_dcat_ap_no_uri("hasGeneralConcept"),
                ]

                for predicate in element_ref_predicates:
                    element_refs = resource_list(graph, prop_ref, predicate)
                    info_model = add_elements_to_model(info_model, graph, element_refs)

    return info_model



from dataclasses import dataclass
from typing import Optional

from .skos_concept import SkosConcept


@dataclass
class Channel:
    uri: Optional[str] = None
    identifier: Optional[str] = None
    type: Optional[SkosConcept] = None



from typing import Dict, List, Optional

from fdk_rdf_parser.classes import DataService, SkosCode
from .reference_data import DataServiceReferenceData


def extend_data_service_with_reference_data(
    data_service: DataService, ref_data: DataServiceReferenceData
) -> DataService:
    data_service.mediaType = extend_media_types(
        data_service.mediaType, ref_data.media_types
    )

    return data_service


def extend_media_types(
    media_types: Optional[List[SkosCode]], references: Optional[Dict[str, SkosCode]]
) -> Optional[List[SkosCode]]:
    extended = []
    if media_types and references:
        for media_type in media_types:
            reference = find_corresponding_reference(media_type.uri, references)
            if reference:
                extended.append(reference)
    return extended if len(extended) > 0 else None


def find_corresponding_reference(
    media_type: Optional[str], references: Optional[Dict[str, SkosCode]]
) -> Optional[SkosCode]:
    if media_type and references:
        for key in references:
            if key in media_type:
                return references[key]
    return None



from dataclasses import dataclass
from typing import Dict, List, Optional, Set

from .contactpoint import ContactPoint
from .publisher import Publisher
from .skos_code import SkosCode
from .theme import ThemeEU


@dataclass
class PartialDcatResource:
    identifier: Optional[Set[str]] = None
    publisher: Optional[Publisher] = None
    title: Optional[Dict[str, str]] = None
    description: Optional[Dict[str, str]] = None
    descriptionFormatted: Optional[Dict[str, str]] = None
    uri: Optional[str] = None
    accessRights: Optional[SkosCode] = None
    theme: Optional[List[ThemeEU]] = None
    keyword: Optional[List[Dict[str, str]]] = None
    contactPoint: Optional[List[ContactPoint]] = None
    dctType: Optional[str] = None
    issued: Optional[str] = None
    modified: Optional[str] = None
    landingPage: Optional[Set[str]] = None
    language: Optional[List[SkosCode]] = None



from typing import Any, List, Optional

from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import Temporal
from fdk_rdf_parser.rdf_utils import (
    date_value,
    dcat_uri,
    owl_time_uri,
    resource_list,
    schema_uri,
)


def extract_temporal(graph: Graph, subject: URIRef) -> Optional[List[Temporal]]:
    values = []
    for temporal_resource in resource_list(graph, subject, DCTERMS.temporal):
        temporal_uri = None
        if isinstance(temporal_resource, URIRef):
            temporal_uri = temporal_resource.toPython()

        start_value = date_value(graph, temporal_resource, dcat_uri("startDate"))
        end_value = date_value(graph, temporal_resource, dcat_uri("endDate"))

        start_value = (
            date_value(graph, temporal_resource, schema_uri("startDate"))
            if start_value is None
            else start_value
        )
        end_value = (
            date_value(graph, temporal_resource, schema_uri("endDate"))
            if end_value is None
            else end_value
        )

        start_value = (
            extract_owl_time_start(graph, temporal_resource)
            if start_value is None
            else start_value
        )
        end_value = (
            extract_owl_time_end(graph, temporal_resource)
            if end_value is None
            else end_value
        )

        values.append(
            Temporal(uri=temporal_uri, startDate=start_value, endDate=end_value)
        )

    return values if len(values) > 0 else None


def extract_owl_time_instant(graph: Graph, subject: Any) -> Optional[str]:
    owl_date = date_value(graph, subject, owl_time_uri("inXSDDateTime"))
    owl_date = (
        date_value(graph, subject, owl_time_uri("inXSDDateTimeStamp"))
        if owl_date is None
        else owl_date
    )
    owl_date = (
        date_value(graph, subject, owl_time_uri("inXSDDate"))
        if owl_date is None
        else owl_date
    )

    return owl_date


def extract_owl_time_start(graph: Graph, subject: Any) -> Any:
    has_beginning = graph.value(subject, owl_time_uri("hasBeginning"))
    if has_beginning is not None:
        return extract_owl_time_instant(graph, has_beginning)
    else:
        return None


def extract_owl_time_end(graph: Graph, subject: Any) -> Any:
    has_end = graph.value(subject, owl_time_uri("hasEnd"))
    if has_end is not None:
        return extract_owl_time_instant(graph, has_end)
    else:
        return None



from typing import List, Optional

from rdflib import Graph, URIRef

from fdk_rdf_parser.classes import SchemaContactPoint
from fdk_rdf_parser.rdf_utils import (
    cv_uri,
    object_value,
    resource_list,
    schema_uri,
    value_translations,
)


def extract_schema_contact_points(
    graph: Graph, subject: URIRef
) -> Optional[List[SchemaContactPoint]]:
    values = []
    for resource in resource_list(graph, subject, cv_uri("hasContactPoint")):
        resource_uri = resource.toPython() if isinstance(resource, URIRef) else None
        values.append(
            SchemaContactPoint(
                uri=resource_uri,
                contactType=value_translations(
                    graph, resource, schema_uri("contactType")
                ),
                description=value_translations(
                    graph, resource, schema_uri("description")
                ),
                email=object_value(graph, resource, schema_uri("email")),
                name=value_translations(graph, resource, schema_uri("name")),
                telephone=object_value(graph, resource, schema_uri("telephone")),
                url=object_value(graph, resource, schema_uri("url")),
            )
        )

    return values if len(values) > 0 else None



from typing import List, Optional

from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import Rule
from fdk_rdf_parser.rdf_utils import (
    cpsv_uri,
    object_value,
    resource_list,
    value_translations,
)
from .dcat_resource import extract_skos_code_list


def extract_rules(graph: Graph, subject: URIRef) -> Optional[List[Rule]]:
    values = []
    for resource in resource_list(graph, subject, cpsv_uri("follows")):
        resource_uri = resource.toPython() if isinstance(resource, URIRef) else None
        values.append(
            Rule(
                uri=resource_uri,
                identifier=object_value(graph, resource, DCTERMS.identifier),
                description=value_translations(graph, resource, DCTERMS.description),
                language=extract_skos_code_list(graph, resource, DCTERMS.language),
                name=value_translations(graph, resource, DCTERMS.name),
            )
        )

    return values if len(values) > 0 else None



from typing import Any, List, Optional

from rdflib import Graph, URIRef

from fdk_rdf_parser.classes import ContactPoint
from fdk_rdf_parser.rdf_utils import dcat_uri, object_value, resource_list, vcard_uri


def extract_contact_points(
    graph: Graph, subject: URIRef
) -> Optional[List[ContactPoint]]:
    values = []
    for resource in resource_list(graph, subject, dcat_uri("contactPoint")):
        resource_uri = None
        if isinstance(resource, URIRef):
            resource_uri = resource.toPython()

        values.append(
            ContactPoint(
                uri=resource_uri,
                fullname=object_value(graph, resource, vcard_uri("fn")),
                email=extract_has_email(graph, resource),
                organizationName=object_value(
                    graph, resource, vcard_uri("hasOrganizationName")
                ),
                organizationUnit=object_value(
                    graph, resource, vcard_uri("organization-unit")
                ),
                hasURL=object_value(graph, resource, vcard_uri("hasURL")),
                hasTelephone=extract_has_telephone(graph, resource),
            )
        )

    return values if len(values) > 0 else None


def extract_has_telephone(graph: Graph, subject: Any) -> Optional[Any]:
    tel_ref = graph.value(subject, vcard_uri("hasTelephone"))
    if tel_ref:
        tel_value = object_value(graph, tel_ref, vcard_uri("hasValue"))
        if tel_value is None:
            tel_value = tel_ref.toPython()

        if "tel:" in tel_value:
            return tel_value[4:]
        else:
            return tel_value
    else:
        return None


def extract_has_email(graph: Graph, subject: Any) -> Optional[Any]:
    mail_ref = graph.value(subject, vcard_uri("hasEmail"))
    if mail_ref:
        email = object_value(graph, mail_ref, vcard_uri("hasValue"))
        if email is None:
            email = mail_ref.toPython()

        if "mailto:" in email:
            return email[7:]
        else:
            return email
    else:
        return None



from dataclasses import dataclass
from typing import Dict, Optional

from .publisher import Publisher


@dataclass
class Catalog:
    id: Optional[str] = None
    publisher: Optional[Publisher] = None
    title: Optional[Dict[str, str]] = None
    uri: Optional[str] = None
    description: Optional[Dict[str, str]] = None



from dataclasses import dataclass
from typing import Dict, List, Optional

from .agent import Agent
from .skos_concept import SkosConcept


@dataclass
class Participation:
    uri: Optional[str] = None
    identifier: Optional[str] = None
    description: Optional[Dict[str, str]] = None
    role: Optional[List[SkosConcept]] = None
    agents: Optional[List[Agent]] = None



from dataclasses import dataclass
from typing import Dict, List, Optional, Set

from .contactpoint import ContactPoint
from .harvest_meta_data import HarvestMetaData
from .publisher import Publisher


@dataclass
class Collection:
    id: Optional[str] = None
    uri: Optional[str] = None
    label: Optional[Dict[str, str]] = None
    description: Optional[Dict[str, str]] = None
    publisher: Optional[Publisher] = None


@dataclass
class TextAndURI:
    uri: Optional[str] = None
    text: Optional[Dict[str, str]] = None


@dataclass
class Definition:
    text: Optional[Dict[str, str]] = None
    remark: Optional[Dict[str, str]] = None
    targetGroup: Optional[str] = None
    sourceRelationship: Optional[str] = None
    range: Optional[TextAndURI] = None
    sources: Optional[List[TextAndURI]] = None
    lastUpdated: Optional[str] = None


@dataclass
class Concept:
    id: Optional[str] = None
    uri: Optional[str] = None
    harvest: Optional[HarvestMetaData] = None
    collection: Optional[Collection] = None
    publisher: Optional[Publisher] = None
    subject: Optional[Dict[str, str]] = None
    application: Optional[List[Dict[str, str]]] = None
    example: Optional[Dict[str, str]] = None
    prefLabel: Optional[Dict[str, str]] = None
    hiddenLabel: Optional[List[Dict[str, str]]] = None
    altLabel: Optional[List[Dict[str, str]]] = None
    contactPoint: Optional[List[ContactPoint]] = None
    definition: Optional[List[Definition]] = None
    seeAlso: Optional[Set[str]] = None
    validFromIncluding: Optional[str] = None
    validToIncluding: Optional[str] = None
    type: str = "concept"



from typing import Dict, Optional

from rdflib import Graph, URIRef
from rdflib.namespace import RDF

from fdk_rdf_parser.classes import QualityAnnotation
from fdk_rdf_parser.rdf_utils import (
    dqv_uri,
    object_value,
    prov_uri,
    resource_list,
    value_translations,
)


def extract_quality_annotation(
    graph: Graph, subject: URIRef
) -> Dict[str, QualityAnnotation]:
    annotations = {}

    for annotation in resource_list(graph, subject, dqv_uri("hasQualityAnnotation")):
        in_dimension_value = object_value(graph, annotation, dqv_uri("inDimension"))
        if in_dimension_value is not None and isinstance(in_dimension_value, str):
            annotations[in_dimension_value] = QualityAnnotation(
                inDimension=in_dimension_value,
                hasBody=extract_has_body(graph, annotation),
            )

    return annotations


def extract_has_body(graph: Graph, subject: URIRef) -> Optional[Dict[str, str]]:
    has_body_uri = graph.value(subject, prov_uri("hasBody"))
    if has_body_uri is not None:
        return value_translations(graph, has_body_uri, RDF.value)
    else:
        return None



from dataclasses import dataclass
from typing import Dict, Optional


@dataclass
class SchemaContactPoint:
    uri: Optional[str] = None
    contactType: Optional[Dict[str, str]] = None
    description: Optional[Dict[str, str]] = None
    email: Optional[str] = None
    name: Optional[Dict[str, str]] = None
    telephone: Optional[str] = None
    url: Optional[str] = None
